@@ -1,12 +1,12 @@
 ;; -*- no-byte-compile: t; lexical-binding: nil -*-
-(define-package "helm-system-packages" "1.10.2"
+(define-package "helm-system-packages" "20251119.751"
   "Helm UI wrapper for system package managers."
   '((emacs "24.4")
     (helm  "2.8.7")
     (seq   "1.8"))
   :url "https://github.com/emacs-helm/helm-system-packages"
-  :commit "c331c69de0a37d2bc4d6f882cc021a905e7e56f9"
-  :revdesc "c331c69de0a3"
+  :commit "002c40e14714a1d3e87c340410f2454d7df681b6"
+  :revdesc "002c40e14714"
   :keywords '("helm" "packages")
   :authors '(("Pierre Neidhardt" . "mail@ambrevar.xyz"))
   :maintainers '(("Pierre Neidhardt" . "mail@ambrevar.xyz")))
